from bitshares_signing.types import (Array, Extension, Id, Int64, ObjectId,
                                     Optional, PointInTime, Signature,
                                     StaticVariant, String, Uint8, Uint16,
                                     Uint32)


def is_args_this_class(self, args):
    """
    graphenebase/objects.py
    True if there is only one argument
    and its type name is the same as the type name of self
    """
    return len(args) == 1 and type(args[0]).__name__ == type(self).__name__


# SERIALIZATION OBJECTS
class GrapheneObject:
    def __init__(self, *args, **kwargs):
        if is_args_this_class(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            self.data = self._prepare_data(kwargs)

    def _prepare_data(self, kwargs):
        """To be implemented by subclasses. Preprocess kwargs and return the data dict."""
        raise NotImplementedError("Subclasses must implement _prepare_data")

    def __bytes__(self):
        # encodes data into wire format
        if self.data is None:
            return bytes()
        b = b""
        # data is a dictionary for human readable reasons
        # but is treated as an ordered list of values
        for value in self.data.values():
            b += bytes(value, "utf-8") if isinstance(value, str) else bytes(value)
        return b


class Price(GrapheneObject):
    def _prepare_data(self, kwargs):
        return {"base": Asset(kwargs["base"]), "quote": Asset(kwargs["quote"])}


class Asset(GrapheneObject):  # bitsharesbase/objects.py
    def _prepare_data(self, kwargs):
        return {
            "amount": Int64(kwargs["amount"]),
            "asset_id": ObjectId(kwargs["asset_id"], "asset"),
        }


class PriceFeed(GrapheneObject):
    """
    bitsharesbase/objects.py
    """

    def _prepare_data(self, kwargs):
        return {
            "settlement_price": Price(kwargs["settlement_price"]),
            "maintenance_collateral_ratio": Uint16(
                kwargs["maintenance_collateral_ratio"]
            ),
            "maximum_short_squeeze_ratio": Uint16(
                kwargs["maximum_short_squeeze_ratio"]
            ),
            "core_exchange_rate": Price(kwargs["core_exchange_rate"]),
        }


class Asset_create(GrapheneObject):
    def _prepare_data(self, kwargs):
        if kwargs.get("bitasset_opts"):
            bitasset_opts = Optional(BitAssetOptions(kwargs["bitasset_opts"]))
        else:
            bitasset_opts = Optional(None)
        return {
            "fee": Asset(kwargs["fee"]),
            "issuer": ObjectId(kwargs["issuer"], "account"),
            "symbol": String(kwargs["symbol"]),
            "precision": Uint8(kwargs["precision"]),
            "common_options": AssetOptions(kwargs["common_options"]),
            "bitasset_opts": bitasset_opts,
            "is_prediction_market": Uint8(bool(kwargs["is_prediction_market"])),
            "extensions": Array([]),
        }


class AssetOptions(GrapheneObject):
    def _prepare_data(self, kwargs):
        return {
            "max_supply": Int64(kwargs["max_supply"]),
            "market_fee_percent": Uint16(kwargs["market_fee_percent"]),
            "max_market_fee": Int64(kwargs["max_market_fee"]),
            "issuer_permissions": Uint16(kwargs["issuer_permissions"]),
            "flags": Uint16(kwargs["flags"]),
            "core_exchange_rate": Price(kwargs["core_exchange_rate"]),
            "whitelist_authorities": Array(
                [ObjectId(x, "account") for x in kwargs["whitelist_authorities"]]
            ),
            "blacklist_authorities": Array(
                [ObjectId(x, "account") for x in kwargs["blacklist_authorities"]]
            ),
            "whitelist_markets": Array(
                [ObjectId(x, "asset") for x in kwargs["whitelist_markets"]]
            ),
            "blacklist_markets": Array(
                [ObjectId(x, "asset") for x in kwargs["blacklist_markets"]]
            ),
            "description": String(kwargs["description"]),
            "extensions": Array([]),
        }


class Asset_reserve(GrapheneObject):
    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "payer": ObjectId(kwargs["payer"], "account"),
            "amount_to_reserve": Asset(kwargs["amount_to_reserve"]),
            "extensions": Array([]),
        }


class BitAssetOptions(GrapheneObject):
    def _prepare_data(self, kwargs):
        return {
            "feed_lifetime_sec": Uint32(kwargs["feed_lifetime_sec"]),
            "minimum_feeds": Uint8(kwargs["minimum_feeds"]),
            "force_settlement_delay_sec": Uint32(kwargs["force_settlement_delay_sec"]),
            "force_settlement_offset_percent": Uint16(
                kwargs["force_settlement_offset_percent"]
            ),
            "maximum_force_settlement_volume": Uint16(
                kwargs["maximum_force_settlement_volume"]
            ),
            "short_backing_asset": ObjectId(kwargs["short_backing_asset"], "asset"),
            "extensions": Array([]),
        }


class Asset_claim_pool(GrapheneObject):
    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "issuer": ObjectId(kwargs["issuer"], "account"),
            "asset_id": ObjectId(kwargs["asset_id"], "asset"),
            "amount_to_claim": Asset(kwargs["amount_to_claim"]),
            "extensions": Array([]),
        }


class Call_order_update(GrapheneObject):
    """
    /bitsharesbase/operations.py
    """

    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "funding_account": ObjectId(kwargs["funding_account"], "account"),
            "delta_collateral": Asset(kwargs["delta_collateral"]),
            "delta_debt": Asset(kwargs["delta_debt"]),
            "extensions": CallOrderExtension(kwargs["extensions"]),
        }


class Limit_order_create(GrapheneObject):  # bitsharesbase/operations.py
    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "seller": ObjectId(kwargs["seller"], "account"),
            "amount_to_sell": Asset(kwargs["amount_to_sell"]),
            "min_to_receive": Asset(kwargs["min_to_receive"]),
            "expiration": PointInTime(kwargs["expiration"]),
            "fill_or_kill": Uint8(kwargs["fill_or_kill"]),
            "extensions": Array([]),
        }


class Limit_order_cancel(GrapheneObject):  # bitsharesbase/operations.py
    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "fee_paying_account": ObjectId(kwargs["fee_paying_account"], "account"),
            "order": ObjectId(kwargs["order"], "limit_order"),
            "extensions": Array([]),
        }


class CallOrderExtension(Extension):
    """
    /bitsharesbase/objects.py
    """

    # FIXME this "should" have self as first arg but pybitshares does it like this
    def tcr(value):
        if value:
            return Uint16(value)
        else:
            return Optional(None)

    sorted_options = [("target_collateral_ratio", tcr)]


class Asset_publish_feed(GrapheneObject):
    """
    bitsharesbase/operations.py
    """

    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "publisher": ObjectId(kwargs["publisher"], "account"),
            "asset_id": ObjectId(kwargs["asset_id"], "asset"),
            "feed": PriceFeed(kwargs["feed"]),
            "extensions": Array([]),
        }


class Liquidity_pool_create(GrapheneObject):
    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "account": ObjectId(kwargs["account"], "account"),
            "asset_a": ObjectId(kwargs["asset_a"], "asset"),
            "asset_b": ObjectId(kwargs["asset_b"], "asset"),
            "share_asset": ObjectId(kwargs["share_asset"], "asset"),
            "taker_fee_percent": Uint16(kwargs["taker_fee_percent"]),
            "withdrawal_fee_percent": Uint16(kwargs["withdrawal_fee_percent"]),
            "extensions": Array([]),
        }


class Liquidity_pool_deposit(GrapheneObject):
    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "account": ObjectId(kwargs["account"], "account"),
            "pool": ObjectId(kwargs["pool"], "liquidity_pool"),
            "amount_a": Asset(kwargs["amount_a"]),
            "amount_b": Asset(kwargs["amount_b"]),
            "extensions": Array([]),
        }


class Liquidity_pool_exchange(GrapheneObject):  # bitsharesbase/operations.py
    def _prepare_data(self, kwargs):
        return {
            "fee": Asset(kwargs["fee"]),
            "account": ObjectId(kwargs["account"], "account"),
            "pool": ObjectId(kwargs["pool"], "liquidity_pool"),
            "amount_to_sell": Asset(kwargs["amount_to_sell"]),
            "min_to_receive": Asset(kwargs["min_to_receive"]),
            "extensions": Array([]),
        }


class Asset_update_feed_producers(GrapheneObject):
    """
    bitsharesbase/operations.py
    """

    def _prepare_data(self, kwargs):
        kwargs["new_feed_producers"] = sorted(
            kwargs["new_feed_producers"], key=lambda x: float(x.split(".")[2])
        )
        return {
            "fee": Asset(kwargs["fee"]),
            "issuer": ObjectId(kwargs["issuer"], "account"),
            "asset_to_update": ObjectId(kwargs["asset_to_update"], "asset"),
            "new_feed_producers": Array(
                [ObjectId(o, "account") for o in kwargs["new_feed_producers"]]
            ),
            "extensions": Array([]),
        }



r"""
graphene_signing.py

  ____  _ _   ____  _                         
 | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___ 
 |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|
 | |_) | | |_ ___) | | | | (_| | | |  __/\__ \
 |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/
       ____  _             _                  
      / ___|(_) __ _ _ __ (_)_ __   __ _      
      \___ \| |/ _` | '_ \| | '_ \ / _` |     
       ___) | | (_| | | | | | | | | (_| |     
      |____/|_|\__, |_| |_|_|_| |_|\__, |     
               |___/               |___/   


WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

Graphene types and signing function required to authenticate operation being broadcast to BitShares

"""
# DISABLE SELECT PYLINT TESTS

# STANDARD PYTHON MODULES
from binascii import hexlify  # binary text to hexidecimal
from binascii import unhexlify  # hexidecimal to binary text
from hashlib import new as hashlib_new  # access algorithm library
from hashlib import sha256  # message digest algorithm
from json import dumps as json_dumps  # serialize object to string
from json import loads as json_loads  # deserialize string to object
from struct import pack  # convert to string representation of C struct

# GRAPHENE SIGNING MODULES
from bitshares_signing.config import ID, PREFIX
# if there was ever a use for "import *"...
from bitshares_signing.operations import (Asset, Asset_claim_pool,
                                          Asset_create, Asset_publish_feed,
                                          Asset_reserve,
                                          Asset_update_feed_producers,
                                          AssetOptions, BitAssetOptions,
                                          Call_order_update,
                                          CallOrderExtension, GrapheneObject,
                                          Limit_order_cancel,
                                          Limit_order_create,
                                          Liquidity_pool_create,
                                          Liquidity_pool_deposit,
                                          Liquidity_pool_exchange, Price,
                                          PriceFeed)
from bitshares_signing.rpc import rpc_get_transaction_hex
from bitshares_signing.types import (Array, Bytes, Id, ObjectId, Optional,
                                     PointInTime, Signature, Uint16, Uint32,
                                     Uint64, varint)
from bitshares_signing.utilities import from_iso_date, it
# THIRD PARTY MODULES
from ecdsa import SECP256k1 as ecdsa_SECP256k1  # curve
from ecdsa import SigningKey as ecdsa_SigningKey  # class
from ecdsa import VerifyingKey as ecdsa_VerifyingKey  # class
from ecdsa import numbertheory as ecdsa_numbertheory  # largest import
from ecdsa import util as ecdsa_util  # module
from secp256k1 import PrivateKey as secp256k1_PrivateKey  # class
from secp256k1 import PublicKey as secp256k1_PublicKey  # class
from secp256k1 import ffi as secp256k1_ffi  # compiled ffi object
from secp256k1 import lib as secp256k1_lib  # library

# GLOBAL CONSTANTS
# base58 encoding and decoding
# this is alphabet defined minus similar characters like 0 and O:
BASE58 = b"123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
# hex encoding and decoding
HEXDIGITS = "0123456789abcdefABCDEF"
ALL_FLAGS = (
    secp256k1_lib.SECP256K1_CONTEXT_VERIFY | secp256k1_lib.SECP256K1_CONTEXT_SIGN
)


# BASE 58 ENCODE, DECODE, AND CHECK "  # graphenebase/base58.py
class Base58:
    """
    This class serves as an abstraction layer
    to deal with base58 encoded strings
    and their corresponding hex and binary representation
    """

    def __init__(self, data, prefix=PREFIX):
        self._prefix = prefix
        if all(c in HEXDIGITS for c in data):
            self._hex = data
        elif data[0] in ["5", "6"]:
            self._hex = base58_check_decode(data)
        elif data[0] in ["K", "L"]:
            self._hex = base58_check_decode(data)[:-2]
        elif data[: len(self._prefix)] == self._prefix:
            self._hex = gph_base58_check_decode(data[len(self._prefix) :])
        else:
            raise ValueError("Error loading Base58 object")

    def __format__(self, _format):
        assert _format.upper() == PREFIX
        return _format.upper() + str(self)

    def __repr__(self):  # hex string of data
        return self._hex

    def __str__(self):  # base58 string of data
        return gph_base58_check_encode(self._hex)

    def __bytes__(self):  # raw bytes of data
        return unhexlify(self._hex)


def base58_decode(base58_str):
    """
    #
    """
    base58_text = bytes(base58_str, "ascii")
    num = 0
    leading_zeroes_count = 0
    for byte in base58_text:
        num = num * 58 + BASE58.find(byte)
        if num == 0:
            leading_zeroes_count += 1
    res = bytearray()
    while num >= 256:
        div, mod = divmod(num, 256)
        res.insert(0, mod)
        num = div
    res.insert(0, num)
    return hexlify(bytearray(1) * leading_zeroes_count + res).decode("ascii")


def base58_encode(hexstring):
    """
    #
    """
    byteseq = bytes(unhexlify(bytes(hexstring, "ascii")))
    num = 0
    leading_zeroes_count = 0
    for byte in byteseq:
        num = num * 256 + byte
        if num == 0:
            leading_zeroes_count += 1
    res = bytearray()
    while num >= 58:
        div, mod = divmod(num, 58)
        res.insert(0, BASE58[mod])
        num = div
    res.insert(0, BASE58[num])
    return (BASE58[0:1] * leading_zeroes_count + res).decode("ascii")


def ripemd160(string):
    """
    160-bit cryptographic hash function
    """
    r160 = hashlib_new("ripemd160")  # import the library
    r160.update(unhexlify(string))
    return r160.digest()


def double_sha256(string):
    """
    double sha256 cryptographic hash function
    """
    return sha256(sha256(unhexlify(string)).digest()).digest()


def base58_check_encode(version, payload):
    """
    #
    """
    string = ("%.2x" % version) + payload
    checksum = double_sha256(string)[:4]
    return base58_encode(string + hexlify(checksum).decode("ascii"))


def gph_base58_check_encode(string):
    """
    #
    """
    checksum = ripemd160(string)[:4]
    return base58_encode(string + hexlify(checksum).decode("ascii"))


def base58_check_decode(string):
    """
    #
    """
    string = unhexlify(base58_decode(string))
    dec = hexlify(string[:-4]).decode("ascii")
    checksum = double_sha256(dec)[:4]
    assert string[-4:] == checksum
    return dec[2:]


def gph_base58_check_decode(string):
    """
    #
    """
    string = unhexlify(base58_decode(string))
    dec = hexlify(string[:-4]).decode("ascii")
    checksum = ripemd160(dec)[:4]
    assert string[-4:] == checksum
    return dec


# ADDRESS AND KEYS
class Address:
    """
    # cropped litepresence2019
    Example :: Address("BTSFN9r6VYzBK8EKtMewfNbfiGCr56pHDBFi")
    # graphenebase/account.py
    """

    def __init__(self, address=None, pubkey=None, prefix=PREFIX):
        self.prefix = prefix
        self.pubkey = Base58(pubkey, prefix=prefix)
        self._address = address


class PublicKey:
    """
    # graphenebase/account.py
    This class deals with Public Keys and inherits ``Address``.
    :param str pk: Base58 encoded public key
    :param str prefix: Network prefix (defaults to ``BTS``)
    """

    def __init__(self, pk, prefix=PREFIX):
        self.prefix = prefix
        self._pk = Base58(pk, prefix=prefix)
        self.address = Address(pubkey=pk, prefix=prefix)
        self.pubkey = self._pk

    def derive_y_from_x(self, xxx, is_even):
        """
        Derive yyy point from xxx point:
        e: yyy^2 = xxx^3 + ax + bbb mod ppp
        """
        curve = ecdsa_SECP256k1.curve
        aaa, bbb, ppp = curve.a(), curve.b(), curve.p()
        alpha = (pow(xxx, 3, ppp) + aaa * xxx + bbb) % ppp
        beta = ecdsa_numbertheory.square_root_mod_prime(alpha, ppp)
        if (beta % 2) == is_even:
            beta = ppp - beta
        return beta

    def compressed(self):
        """
        Derive compressed public key
        """
        order = ecdsa_SECP256k1.generator.order()
        point = ecdsa_VerifyingKey.from_string(
            bytes(self), curve=ecdsa_SECP256k1
        ).pubkey.point
        x_str = ecdsa_util.number_to_string(point.x(), order)
        # y_str = ecdsa_util.number_to_string(point.y(), order)
        compressed = hexlify(bytes(chr(2 + (point.y() & 1)), "ascii") + x_str).decode(
            "ascii"
        )
        return compressed

    def un_compressed(self):
        """
        Derive uncompressed key
        """
        public_key = repr(self._pk)
        prefix = public_key[0:2]
        if prefix == "04":
            return public_key
        assert prefix in ["02", "03"]
        xxx = int(public_key[2:], 16)
        yyy = self.derive_y_from_x(xxx, (prefix == "02"))
        return "04" + "%064x" % xxx + "%064x" % yyy

    def point(self):
        """
        Return the point for the public key
        """
        string = unhexlify(self.un_compressed())
        return ecdsa_VerifyingKey.from_string(
            string[1:], curve=ecdsa_SECP256k1
        ).pubkey.point

    def __repr__(self):
        """
        Gives the hex representation of the Graphene public key.
        """
        return repr(self._pk)

    def __format__(self, _format):
        """
        Formats the instance of:doc:`Base58 <base58>
        ` according to ``_format``
        """
        return format(self._pk, _format)

    def __bytes__(self):
        """
        Returns the raw public key (has length 33)
        """
        return bytes(self._pk)


class PrivateKey:
    """
    Derives the compressed and uncompressed public keys and
    constructs two instances of ``PublicKey``:
    # Bitshares(MIT) graphenebase/account.py
    # Bitshares(MIT) bitsharesbase/account.py
    # merged litepresence2019
    """

    def __init__(self, wif=None, prefix=PREFIX):
        self._wif = wif if isinstance(wif, Base58) else Base58(wif)
        self._pubkeyhex, self._pubkeyuncompressedhex = self.compressed_pubkey()
        self.pubkey = PublicKey(self._pubkeyhex, prefix=prefix)
        self.uncompressed = PublicKey(self._pubkeyuncompressedhex, prefix=prefix)
        self.uncompressed.address = Address(
            pubkey=self._pubkeyuncompressedhex, prefix=prefix
        )
        self.address = Address(pubkey=self._pubkeyhex, prefix=prefix)

    def compressed_pubkey(self):
        """
        Derive uncompressed public key
        """
        secret = unhexlify(repr(self._wif))
        order = ecdsa_SigningKey.from_string(
            secret, curve=ecdsa_SECP256k1
        ).curve.generator.order()
        point = ecdsa_SigningKey.from_string(
            secret, curve=ecdsa_SECP256k1
        ).verifying_key.pubkey.point
        x_str = ecdsa_util.number_to_string(point.x(), order)
        y_str = ecdsa_util.number_to_string(point.y(), order)
        compressed = hexlify(chr(2 + (point.y() & 1)).encode("ascii") + x_str).decode(
            "ascii"
        )
        uncompressed = hexlify(chr(4).encode("ascii") + x_str + y_str).decode("ascii")
        return [compressed, uncompressed]

    def __bytes__(self):
        """
        Returns the raw private key
        """
        return bytes(self._wif)


class Memo(GrapheneObject):
    def _prepare_data(self, kwargs):
        prefix = kwargs.pop("prefix", PREFIX)
        if "message" in kwargs and kwargs["message"]:
            return {
                "from": PublicKey(kwargs["from"], prefix=prefix),
                "to": PublicKey(kwargs["to"], prefix=prefix),
                "nonce": Uint64(int(kwargs["nonce"])),
                "message": Bytes(kwargs["message"]),
            }
        else:
            return None


class Transfer(GrapheneObject):
    def _prepare_data(self, kwargs):
        prefix = kwargs.get("prefix", PREFIX)
        if "memo" in kwargs and kwargs["memo"]:
            if isinstance(kwargs["memo"], dict):
                kwargs["memo"]["prefix"] = prefix
                memo = Optional(Memo(**kwargs["memo"]))
            else:
                memo = Optional(Memo(kwargs["memo"]))
        else:
            memo = Optional(None)
        return {
            "fee": Asset(kwargs["fee"]),
            "from": ObjectId(kwargs["from"], "account"),
            "to": ObjectId(kwargs["to"], "account"),
            "amount": Asset(kwargs["amount"]),
            "memo": memo,
            "extensions": Array([]),
        }


class Asset_issue(GrapheneObject):
    def _prepare_data(self, kwargs):
        prefix = kwargs.get("prefix", PREFIX)
        if "memo" in kwargs and kwargs["memo"]:
            memo = Optional(Memo(prefix=prefix, **kwargs["memo"]))  # FIXME
        else:
            memo = Optional(None)
        return {
            "fee": Asset(kwargs["fee"]),
            "issuer": ObjectId(kwargs["issuer"], "account"),
            "asset_to_issue": Asset(kwargs["asset_to_issue"]),
            "issue_to_account": ObjectId(kwargs["issue_to_account"], "account"),
            "memo": memo,
            "extensions": Array([]),
        }


class Operation:  # refactored  litepresence2019
    "class GPHOperation():"
    # Bitshares(MIT) graphenebase/objects.py
    "class Operation(GPHOperation):"
    # Bitshares(MIT) bitsharesbase/objects.py

    def __init__(self, op):
        if not isinstance(op, list):
            raise ValueError("expecting op to be a list")
        if len(op) != 2:
            raise ValueError("expecting op to be two items")
        if not isinstance(op[0], int):
            raise ValueError("expecting op[0] to be integer")

        self.opId = op[0]

        # Define a mapping of operation codes to their corresponding classes
        operation_map = {
            0: Transfer,
            1: Limit_order_create,
            2: Limit_order_cancel,
            3: Call_order_update,
            10: Asset_create,
            13: Asset_update_feed_producers,
            14: Asset_issue,
            15: Asset_reserve,
            19: Asset_publish_feed,
            47: Asset_claim_pool,
            59: Liquidity_pool_create,
            61: Liquidity_pool_deposit,
            63: Liquidity_pool_exchange,
        }

        # Use the mapping to set self.op
        if op[0] in operation_map:
            self.op = operation_map[op[0]](op[1])
        else:
            raise ValueError(f"Invalid operation code: {op[0]}")

    def __bytes__(self):
        print(it("yellow", "GPHOperation.__bytes__"))
        return bytes(Id(self.opId)) + bytes(self.op)


class SignedTransaction(GrapheneObject):  # merged litepresence2019
    # Bitshares(MIT) graphenebase/signedtransactions.py
    # Bitshares(MIT) bitsharesbase/signedtransactions.py

    def _prepare_data(self, kwargs):
        print(it("red", "SignedTransaction"))
        print(
            """ Create a signed transaction and
                    offer method to create the signature

                (see ``getBlockParams``)
                :param num refNum: parameter ref_block_num
                :param num refPrefix: parameter ref_block_prefix
                :param str expiration: expiration date
                :param Array operations:  array of operations
            """
        )
        print("kwargs", kwargs)
        if (
            "extensions" not in kwargs
            or "extensions" in kwargs
            and not kwargs.get("extensions")
        ):
            kwargs["extensions"] = Array([])
        if "signatures" not in kwargs:
            kwargs["signatures"] = Array([])
        else:
            kwargs["signatures"] = Array(
                [Signature(unhexlify(a)) for a in kwargs["signatures"]]
            )

        if "operations" in kwargs:
            opklass = self.getOperationKlass()
            if all(not isinstance(a, opklass) for a in kwargs["operations"]):
                kwargs["operations"] = Array([opklass(a) for a in kwargs["operations"]])
            else:
                kwargs["operations"] = Array(kwargs["operations"])

        return {
                "ref_block_num": Uint16(kwargs["ref_block_num"]),
                "ref_block_prefix": Uint32(kwargs["ref_block_prefix"]),
                "expiration": PointInTime(kwargs["expiration"]),
                "operations": kwargs["operations"],
                "extensions": kwargs["extensions"],
                "signatures": kwargs["signatures"],
            }

    @property
    def id(self):
        print("SignedTransaction.id")
        """
        The transaction id of this transaction
        """
        # Store signatures temporarily
        sigs = self.data["signatures"]
        self.data.pop("signatures", None)
        # Generage Hash of the seriliazed version
        h = sha256(bytes(self)).digest()
        # Recover signatures
        self.data["signatures"] = sigs
        # Return properly truncated tx hash
        return hexlify(h[:20]).decode("ascii")

    def getOperationKlass(self):
        print("SignedTransaction.get_operationKlass")
        return Operation

    # FIXME ded code?
    # def derSigToHexSig(self, s):
    #     print("SignedTransaction.derSigToHexSig")
    #     s, junk = ecdsa_der.remove_sequence(unhexlify(s))
    #     if junk:
    #         print("JUNK: %s", hexlify(junk).decode("ascii"))
    #     assert junk == b""
    #     x, s = ecdsa_der.remove_integer(s)
    #     y, s = ecdsa_der.remove_integer(s)
    #     return "%064x%064x" % (x, y)

    def derive_digest(self, chain):
        print("SignedTransaction.derive_digest")
        print(self, chain)
        # Do not serialize signatures
        sigs = self.data["signatures"]
        self.data["signatures"] = []
        # Get message to sign
        #   bytes(self) will give the wire formated data according to
        #   GrapheneObject and the data given in __init__()
        self.message = unhexlify(ID) + bytes(self)
        self.digest = sha256(self.message).digest()
        # restore signatures
        self.data["signatures"] = sigs

    def verify(self, pubkeys=[], chain=PREFIX):
        print(it("green", "###############################################"))
        print("SignedTransaction.verify")
        print(it("green", "self, pubkeys, chain"), self, pubkeys, chain)

        self.derive_digest(chain)
        print(it("green", "self"))
        print(self)
        signatures = self.data["signatures"].data
        print(it("green", "signatures"))
        print(signatures)
        pubKeysFound = []

        for signature in signatures:
            p = verify_message(self.message, bytes(signature))
            phex = hexlify(p).decode("ascii")
            print("")
            print("")
            print(it("green", "phex"))
            print(it("green", phex))
            print(it("cyan", "len(phex)"), len(str(phex)))
            print("")
            print("")
            pubKeysFound.append(phex)

        for pubkey in pubkeys:
            print(it("green", "for pubkey in pubkeys:"))
            print(it("green", "************ pubkey ************"))
            print(it("blue", "repr(pubkey)"))
            print(repr(pubkey))

            print(it("cyan", "len(pubkey)"), len(str(pubkey)))
            print("")
            if not isinstance(pubkey, PublicKey):
                raise Exception("Pubkeys must be array of 'PublicKey'")

            k = pubkey.un_compressed()[2:]

            print(it("green", ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"))
            print(it("yellow", "k"))
            print(k)
            print(it("cyan", "len(k)"), len(str(k)))
            print(it("yellow", "pubKeysFound"))
            print(pubKeysFound)
            print(it("cyan", "len(pubKeysFound[0])"), len(pubKeysFound[0]))
            print("")
            print(it("green", ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"))

            # if k not in pubKeysFound and repr(pubkey) not in pubKeysFound:
            #     print(
            #         it(
            #             "blue",
            #             "if k not in pubKeysFound and repr(pubkey) "
            #             + "not in pubKeysFound:",
            #         )
            #     )
            #     k = PublicKey(PublicKey(k).compressed())
            #     f = format(k, PREFIX)  # chain_params["prefix"]) # 'BTS'
            #     print("")
            #     print(it("red", "FIXME"))
            #     raise Exception("Signature for %s missing!" % f)

        return pubKeysFound

    # FIXME: ded code?
    # def sign(self, wifkeys, chain=PREFIX):
    #     print("SignedTransaction.sign")
    #     """
    #     Sign the transaction with the provided private keys.
    #     """
    #     # FIXME is this even used????
    #     self.derive_digest(chain)

    #     # Get Unique private keys
    #     self.privkeys = []
    #     [self.privkeys.append(item) for item in wifkeys if item not in self.privkeys]

    #     # Sign the message with every private key given!
    #     sigs = []
    #     for wif in self.privkeys:
    #         signature = sign_message(self.message, wif)
    #         sigs.append(Signature(signature))

    #     self.data["signatures"] = Array(sigs)
    #     return self


# SERIALIZATION


def serialize_transaction(rpc, trx):
    """
    {"method":"call","params":[
        2,
        "get_required_signatures",
            [{"ref_block_num":0,
            "ref_block_prefix":0,
            "expiration":"1970-01-01T00:00:00",
            "operations":
                [[3,{
                "fee":{"amount":"48260","asset_id":"1.3.0"},
                "funding_account":"1.2.11111111",
                "delta_collateral":{"amount":"1","asset_id":"1.3.5650"},
                "delta_debt":{"amount":"0","asset_id":"1.3.5662"},
                "extensions":{"target_collateral_ratio":2000}}]],
                "extensions":[],"signatures":[]},
                ["BTS6upQe7upQe7upQe7upQe7upQe7upQe7upQe7upQe7upQe7upQe7"]]],
            "id":371}
    """
    if trx["operations"] == []:
        return trx, b""
    # gist.github.com/xeroc/9bda11add796b603d83eb4b41d38532b
    # RPC call for ordered dicts which are dumped by the query
    tx_ops = trx["operations"]
    for idx, _ in enumerate(tx_ops):
        if "memo" in tx_ops[idx][1].keys() and tx_ops[idx][1]["memo"] == "":
            tx_ops[idx][1] = {k: v for k, v in tx_ops[idx][1].items() if k != "memo"}
    rpc_tx = dict(trx)
    rpc_tx["operations"] = tx_ops
    # find out how the backend suggests to serialize the trx
    rpc_tx_hex = rpc_get_transaction_hex(rpc, json_loads(json_dumps(rpc_tx)))
    buf = b""  # create an empty byte string buffer
    # add block number, prefix, and trx expiration to the buffer
    buf += pack("<H", trx["ref_block_num"])  # 2 byte int
    buf += pack("<I", trx["ref_block_prefix"])  # 4 byte int
    buf += pack("<I", from_iso_date(trx["expiration"]))  # 4 byte int
    # add length of operations list to buffer
    buf += bytes(varint(len(trx["operations"])))
    # add the operations list to the buffer in graphene type fashion
    for op in trx["operations"]:
        # print(ops[0])  # Int (1=create, 2=cancel, etc.)
        # print(ops[1])  # OrderedDict of operations
        buf += varint(op[0])
        if op[0] == 0:
            buf += bytes(Transfer(op[1]))
        elif op[0] == 1:
            buf += bytes(Limit_order_create(op[1]))
        elif op[0] == 2:
            buf += bytes(Limit_order_cancel(op[1]))
        elif op[0] == 3:
            buf += bytes(Call_order_update(op[1]))
            buf += bytes(varint(0))
        elif op[0] == 10:
            buf += bytes(Asset_create(op[1]))
        elif op[0] == 13:
            buf += bytes(Asset_update_feed_producers(op[1]))
        elif op[0] == 14:
            buf += bytes(Asset_issue(op[1]))
        elif op[0] == 15:
            buf += bytes(Asset_reserve(op[1]))
        elif op[0] == 19:
            buf += bytes(Asset_publish_feed(op[1]))
        elif op[0] == 47:
            buf += bytes(Asset_claim_pool(op[1]))
        elif op[0] == 59:
            buf += bytes(Liquidity_pool_create(op[1]))
        elif op[0] == 61:
            buf += bytes(Liquidity_pool_deposit(op[1]))
        elif op[0] == 63:
            buf += bytes(Liquidity_pool_exchange(op[1]))
    # add legth of (empty) extensions list to buffer
    buf += bytes(varint(len(trx["extensions"])))  # usually, effectively varint(0)
    # this the final manual transaction hex, which should match rpc
    manual_tx_hex = hexlify(buf)
    # prepend the chain ID to the buffer to create final serialized msg
    message = unhexlify(ID) + buf
    # if serialization is correct: rpc_tx_hex = manual_tx_hex plus an empty signature
    assert rpc_tx_hex == manual_tx_hex + b"00", "Serialization Failed"
    return trx, message


def sign_transaction(trx, message, wif):
    """
    # graphenebase/ecdsa.py
    # tools.ietf.org/html/rfc6979
    # @xeroc/steem-transaction-signing-in-a-nutshell
    # @dantheman/steem-and-bitshares-cryptographic-security-update
    # deterministic signatures retain the cryptographic
    # security features associated with digital signatures
    # but can be more easily implemented
    # since they do not need high-quality randomness
    """

    def canonical(sig):
        """
        1 in 4 signatures are randomly canonical; "normal form"
        using the other three causes vulnerability to maleability attacks
        as a metaphor; "require reduced fractions in simplest terms"
        note: 0x80 hex = 10000000 binary = 128 integer
        :return bool():
        """
        sig = bytearray(sig)
        return not any(
            [
                int(sig[0]) & 0x80,
                int(sig[32]) & 0x80,
                sig[0] == 0 and not int(sig[1]) & 0x80,
                sig[32] == 0 and not int(sig[33]) & 0x80,
            ]
        )

    # create fixed length representation of arbitrary length data
    # this will thoroughly obfuscate and compress the transaction
    # signing large data is computationally expensive and time consuming
    # the hash of the data is a relatively small
    # signing hash is more efficient than signing serialization
    digest = sha256(message).digest()
    # ECDSA
    # eliptical curve digital signature algorithm
    # this is where the real hocus pocus lies
    # all of the ordering, typing, serializing, and digesting
    # culminates with the message meeting the wif
    # begin with the 8 bit string representation of private key
    try:
        private_key = bytes(PrivateKey(wif))
    except:
        return
    # create some arbitrary data used by the nonce generation
    ndata = secp256k1_ffi.new("const int *ndata")
    ndata[0] = 0  # it adds "\0x00", then "\0x00\0x00", etc..
    while True:  # repeat process until deterministic and cannonical
        ndata[0] += 1  # increment the arbitrary nonce
        # obtain compiled/binary private key from the wif
        privkey = secp256k1_PrivateKey(private_key, raw=True)
        # create a new recoverable 65 byte ECDSA signature
        sig = secp256k1_ffi.new("secp256k1_ecdsa_recoverable_signature *")
        # parse a compact ECDSA signature (64 bytes + recovery id)
        # returns: 1 = deterministic; 0 = not deterministic
        deterministic = secp256k1_lib.secp256k1_ecdsa_sign_recoverable(
            privkey.ctx,  # initialized context object
            sig,  # array where signature is held
            digest,  # 32-byte message hash being signed
            privkey.private_key,  # 32-byte secret key
            secp256k1_ffi.NULL,  # default nonce function
            ndata,  # incrementing nonce data
        )
        if not deterministic:
            continue
        # we derive the recovery parameter
        # which simplifies the verification of the signature
        # it links the signature to a single unique public key
        # without this parameter, the back-end would need to test
        # for multiple public keys instead of just one
        signature, i = privkey.ecdsa_recoverable_serialize(sig)
        # we ensure that the signature is canonical; simplest/reduced form
        if canonical(signature):
            # add 4 and 27 to stay compatible with other protocols
            i += 4  # compressed
            i += 27  # compact
            # and have now obtained our signature
            break
    # having derived a valid canonical signature
    # we format it in its hexadecimal representation
    # and add it our transactions signatures
    # note that we do not only add the signature
    # but also the recover parameter
    # this kind of signature is then called "compact signature"
    signature = hexlify(pack("<B", i) + signature).decode("ascii")
    trx["signatures"].append(signature)
    return trx


def verify_transaction(trx, wif):
    """
    # gist.github.com/xeroc/9bda11add796b603d83eb4b41d38532b
    # once you have derived your new trx including the signatures
    # verify your transaction and it's signature
    """
    tx2 = SignedTransaction(**trx)
    tx2.derive_digest(PREFIX)
    pubkeys = [PrivateKey(wif).pubkey]
    tx2.verify(pubkeys, PREFIX)
    return trx


def verify_message(message, signature):
    """
    graphenebase/ecdsa.py stripped of non-secp256k1 methods
    returns bytes
    """
    # require message and signature to be bytes
    if not isinstance(message, bytes):
        message = bytes(message, "utf-8")
    if not isinstance(signature, bytes):
        signature = bytes(signature, "utf-8")
    # recover parameter only
    recover_parameter = bytearray(signature)[0] - 4 - 27
    # "bitwise or"; each bit of the output is 0
    # if the corresponding bit of x AND of y is 0, otherwise it's 1

    # ecdsa.PublicKey with additional functions to serialize
    # in uncompressed and compressed formats
    pub = secp256k1_PublicKey(flags=ALL_FLAGS)
    # recover raw signature
    sig = pub.ecdsa_recoverable_deserialize(signature[1:], recover_parameter)
    # recover public key
    verify_pub = secp256k1_PublicKey(pub.ecdsa_recover(message, sig))
    # convert recoverable sig to normal sig
    normal_sig = verify_pub.ecdsa_recoverable_convert(sig)
    # verify
    verify_pub.ecdsa_verify(message, normal_sig)
    ret = verify_pub.serialize(compressed=True)
    return ret



r"""
```
+-------------------------------------------------+
|   ____  _ _   ____  _                           |
|  | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___   |
|  |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|  |
|  | |_) | | |_ ___) | | | | (_| | | |  __/\__ \  |
|  |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/  |
|        ____  _             _                    |
|       / ___|(_) __ _ _ __ (_)_ __   __ _        |
|       \___ \| |/ _` | '_ \| | '_ \ / _` |       |
|        ___) | | (_| | | | | | | | | (_| |       |
|       |____/|_|\__, |_| |_|_|_| |_|\__, |       |
|                |___/               |___/        |
+-------------------------------------------------+
```

WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

** ManualSigning re-imagined **

---------------------------------------------------

## Authenticated BUY/SELL/CANCEL without Pybitshares (MIT) Architecture

```
def WTFPL_v0_March_1765():
    if any([stamps, licenses, taxation, regulation, fiat, etat]):
        try:
            print("No thank you!")
        except:
            return [tar, feathers]
```

### ALPHA RELEASE - PUBLIC DOMAIN, NO WARRANTY

Joe, a CEX algorithmic trader, discovers Bitshares DEX and asks:

"How do I get reliable public API data that never goes stale?"
 - metaNODE.py

"How do I authenticate to the DEX?"
 - manualSIGNING.py

Seven small scripts, totaling a mere 105kb, and DEX algo trading barriers to entry are defeated.

You’re instantly connected and authenticated.

Buy, sell, and cancel orders with seamless connectivity and simple authentication — CEX-style data at your fingertips.

Want more?

For full buy/sell/cancel + order book UI:
 - microDEX.py

For advanced algorithmic trading and backtesting:
 - extinctionEVENT.py

For tracking historical returns on investment:
 - accountBALANCES.py

For the ultimate public API full universe search utility:
 - latencyTEST.py

> NOTE: the above scripts are outdated and may require modification for reliable usage

Whitepapers for deep dives:
 - metaNODEwhitepaper.md
 - manualSIGNINGwhitepaper.md

## FEATURES

- `prototype_order()` generates an order header template.
- Edicts can include any combination of operations.
- Automatically scales buy/sell orders to prevent exceeding account budget.
- Ensure you always have enough funds to cover transaction fees with the last two Bitshares.
- Uses multiprocessing to handle websockets and manage faulty order timeouts.
- New edict `{'op': login}` matches a WIF (Wallet Import Format) to an account name and returns `True`/`False`.
- No dependencies on Pybitshares!

## HOW DO I USE THIS TOOL?

An order is structured as a dictionary of:

`['edicts', 'header', 'nodes']`

See `help(bitshares_signing.quickstart)` for detailed examples.

## OBJECTIVES

- Use only standard Python objects. ✅
- Collect necessary Pybitshares objects (copy, paste, cite). ✅
- Strip unnecessary methods from Pybitshares objects. ✅
- Reorganize classes and definitions in a logical, chronological order. ✅
- Enable users to create orders using simple, human-readable terms. ✅
- Build transactions using Graphene-style terms. ✅
- Serialize transactions. ✅
- Validate serialization using `get_transaction_hex_without_sig()`. ✅
- Sign transactions with ECDSA. ✅
- Validate signed transactions. ✅
- Broadcast transactions to an RPC node. ✅
- Make the script importable as a module and callable with `broker(order)`. ✅
- Allow for a list of buy/sell/cancel actions (edicts). ✅
- Implement `cancel-all` functionality. ✅
- Provide extensive line-by-line commentary for clarity. ✅

## ONGOING:
- Make prototype_order take token- and user- names, rather than ids
- Add Liquidity pool swap/stake/unstake
- Simplify and condense Pybitshares methods. (Still in progress)
- Expand and finalize the `manualSIGNINGwhitepaper.md`. (5200-word rough draft available)
- Transition from object-oriented (class-based) to procedural (function-based) style. (Planned)

## DEPENDENCIES
- Python 3
- Linux recommended, other OSes have not been tested
- ecdsa, secp256k1, and websocket-client. (`pip3 install -r requirements.txt` to get the right versions)

> Note: bitshares_signing has not been tested on python3.10+; for reliable use, use
>       python3.6.9 to python3.9

## LICENSE:
Citations to Pybitshares (MIT) & @xeroc as needed.
Special thanks to @vvk123, @sschiessl, and @harukaff_bot.
All remaining rights under WTFPL March 1765.
"""

from bitshares_signing.graphene_auth import broker, prototype_order

__all__ = ["broker", "prototype_order", "SUPPORTED_OPS", "quickstart"]
SUPPORTED_OPS = [
    "login",
    "buy",
    "sell",
    "cancel",
    "swap",
    "deposit_pool",
    "transfer",
    "issue",
    "reserve",
    "call",
    "fee_pool",
    "create_pool",
    "create_asset",
]


def quickstart():
    """
    from bitshares_signing import broker, prototype_order

    # define order headers
    order = prototype_order(
        {
            "asset_id": "1.3.x",
            "asset_precision": int(),
            "currency_id": "1.3.x",
            "currency_precision": int(),
            "account_id": "1.2.x",
            "account_name": str(),
            "wif": str(),
        }
    )
    test_order1 = order.copy()
    test_order2 = order.copy()

    # test your wif with login
    test_order1["edicts"] = [
        {
            "op": "login",
        }
    ]
    logged_in = broker(order)

    # add edicts
    order["edicts"] = [
        # buy - buying `asset`, receiving `currency`
        {
            "op": "buy",
            "amount": float(),
            "price": float(),
        },
        # sell - selling `asset`, receiving `currency`
        {
            "op": "sell",
            "amount": float(),
            "price": float(),
        },
        # cancel - cancel all `ids` or all open orders if `ids` is ["1.7.X"]
        {
            "op": "cancel",
            "ids": ["1.7.X"],
        },
        # swap - selling `amount` of `currency` for `assets` at `price` on `pool`
        {
            "op": "swap",
            "amount": float(),
            "price": float(),
            "pool": "1.19.x",
        },
        # deposit_pool - deposit `amount_a` of `asset` and `amount_b` of `currency` into `pool`
        {
            "op": "deposit_pool",
            "amount_a": float(),  # asset
            "amount_b": float(),  # currency
            "pool": "1.19.x",
        },
        # transfer - move `amount` of `asset` to `account_id`
        {
            "op": "transfer",
            "amount": float(),
            "account_id": "1.2.x",
        },
        # issue - issue `amount` of `asset` to `account_id`
        {
            "op": "issue",
            "amount": float(),
            "account_id": "1.2.x",
        },
        # reserve - reserve `amount` of `asset`
        {
            "op": "reserve",
            "amount": float(),
        },
        # call - update a call order with a change in debt (`debt_delta`), change in
        #        collateral (`collateral_delta`), and target collateral ratio (`tcr`)
        {
            "op": "call",
            "debt_delta": float(),
            "collateral_delta": float(),
            "tcr": float(),
        },
        # fee_pool - claim `amount_to_claim` from the fee pool of `asset_id`, issued by
        #            `issuer`
        {
            "op": "fee_pool",
            "amount_to_claim": float(),
            "asset_id": "1.3.x",
            "issuer": "1.2.x",
        },
        # create_pool - create a liquidity pool, where asset a and b are the asset and
        #               currency in the header.  The pool token is given by `share_asset`,
        #               and withdrawal and taker fees are given by their respective keys.
        {
            "op": "create_pool",
            "share_asset": "1.3.x",
            "taker_fee_percent": float(),  # these two are in percent, so 0.1 would be amt*0.001
            "withdrawal_fee_percent": float(),
        },
        # create_asset - create a new asset, inline comments given because there are
        #                a lot of values.
        {
            "op": "create_asset",
            "issuer": "1.2.x",  # who is allowed to issue this asset?
            "symbol": str(),  # what is it called?
            "precision": int(),  # how many decimal points can amounts of it be referenced with?
            "common_options": {
                # how much of it can exist?
                "max_supply": int(),
                # the fee paid by the buyer to the issuer during limit orders
                "market_fee_percent": int(),
                # non-changeable number that limits the above value
                "max_market_fee": int(),
                # series of flags which limits the ability of the issuer to continue to edit particular flags
                "issuer_permissions": int(),
                # the ability of the issuer to change capabilities of the UIA such as the ability to transfer
                "flags": int(),
                # the user may pay their market fees using core token at the core exchange rate
                # periodically updated by the issuer from the fee pool maintained by the issuer
                "core_exchange_rate": {
                    "base": {"amount": int(), "asset_id": "1.3.1"},
                    "quote": {"amount": int(), "asset_id": "1.3.0"},
                },
                # the issuer may create a list of only those who have permission to trade this asset
                "whitelist_authorities": list("1.2.x"),
                # as well as those who do not
                "blacklist_authorities": list("1.2.x"),
                # the issuer may create a list of only those other assets this one may be traded against
                "whitelist_markets": list("1.3.x"),
                # as well as those it cannot be
                "blacklist_markets": list("1.3.x"),
                "description": str(),
            },
            "is_prediction_market": bool(),
        },
    ]

    broker(order)
    """
    pass



import json
from binascii import unhexlify
from struct import pack  # convert to string representation of C struct

from bitshares_signing.utilities import from_iso_date

# bitsharesbase/objecttypes.py used by ObjectId() to confirm a.b.c
TYPES = {
    "account": 2,
    "asset": 3,
    "limit_order": 7,
    "liquidity_pool": 19,
}


class ObjectId:
    """
    encodes a.b.c object ids - serializes the *instance* only!
    """

    def __init__(self, object_str, type_verify=None):
        # if after splitting a.b.c there are 3 pieces:
        if len(object_str.split(".")) == 3:
            # assign those three pieces to a, b, and c
            aaa, bbb, ccc = object_str.split(".")
            # assure they are integers
            self.aaa = int(aaa)
            self.bbb = int(bbb)
            self.ccc = int(ccc)
            # serialize the c element; the "instance"
            self.instance = Id(self.ccc)
            self.abc = object_str
            # 1.2.x:account, 1.3.x:asset, 1.7.x:limit, etc as defined in TYPES
            if type_verify:
                assert TYPES[type_verify] == int(bbb), (
                    # except raise error showing mismatch
                    "Object id does not match object type! "
                    + "Excpected %d, got %d" % (TYPES[type_verify], int(bbb))
                )
        else:
            raise Exception("Object id is invalid")

    def __bytes__(self):
        """
        bytes of serialized c element; the "instance"
        """
        return bytes(self.instance)


class Id:
    """
    serializes the c element of "a.b.c" types
    merged with Varint32()
    """

    def __init__(self, data):
        self.data = int(data)

    def __bytes__(self):
        return bytes(varint(self.data))


class Array:
    """
    serializes lists as byte strings
    merged with Set() and Varint32()
    """

    def __init__(self, data):
        self.data = data
        self.length = int(len(self.data))

    def __bytes__(self):
        return bytes(varint(self.length)) + b"".join([bytes(a) for a in self.data])


class Uint8:
    """
    byte string of 8 bit unsigned integers
    merged with Bool()
    """

    def __init__(self, data):
        self.data = int(data)

    def __bytes__(self):
        return pack("<B", self.data)


class Uint16:
    """
    byte string of 16 bit unsigned integers
    """

    def __init__(self, data):
        self.data = int(data)

    def __bytes__(self):
        return pack("<H", self.data)


class Uint32:
    """
    byte string of 32 bit unsigned integers
    """

    def __init__(self, data):
        self.data = int(data)

    def __bytes__(self):
        return pack("<I", self.data)


class Uint64:
    """
    byte string of 64 bit unsigned integers
    """

    def __init__(self, data):
        self.data = int(data)

    def __bytes__(self):
        return pack("<Q", self.data)

class Int64:
    """
    byte string of 64 bit signed integers
    """

    def __init__(self, data):
        self.data = int(data)

    def __bytes__(self):
        return pack("<q", self.data)


def unicodify(data):
    r = []
    for s in data:
        o = ord(s)
        if (o <= 7) or (o == 11) or (o > 13 and o < 32):
            r.append("u%04x" % o)
        elif o == 8:
            r.append("b")
        elif o == 9:
            r.append("\t")
        elif o == 10:
            r.append("\n")
        elif o == 12:
            r.append("f")
        elif o == 13:
            r.append("\r")
        else:
            r.append(s)
    return bytes("".join(r), "utf-8")


class String:
    def __init__(self, d):
        self.data = d

    def __bytes__(self):
        if self.data:
            d = unicodify(self.data)
        else:
            d = b""
        return varint(len(d)) + d

    def __str__(self):
        return "%s" % str(self.data)


class Optional:
    """
    #
    """

    def __init__(self, data):
        self.data = data

    def __bytes__(self):
        if not bool(self.data):
            return bytes(Uint8(0))
        return (
            bytes(Uint8(1)) + bytes(self.data) if bytes(self.data) else bytes(Uint8(0))
        )

    def __str__(self):
        return str(self.data)

    def isempty(self):
        """
        is there no data
        """
        if self.data is None:
            return True
        if not bool(str(self.data)):  # pragma: no cover
            return True
        return not bool(bytes(self.data))


class Signature:
    """
    used to disable bytes() method on Signatures in OrderedDicts
    """

    def __init__(self, data):
        self.data = data

    def __bytes__(self):
        return self.data  # note does NOT return bytes(self.data)


class PointInTime:
    """
    used to pack ISO8601 time as 4 byte unix epoch integer as bytes
    """

    def __init__(self, data):
        self.data = data

    def __bytes__(self):
        return pack("<I", from_iso_date(self.data))


class StaticVariant:
    """
    #
    """

    def __init__(self, data, type_id):
        self.data = data
        self.type_id = type_id

    def __bytes__(self):
        return varint(self.type_id) + bytes(self.data)

    def __str__(self):
        return json.dumps([self.type_id, self.data.json()])

class Bytes:
    """Bytes

    Initializes from and stores internally as a string of hex digits.
    Byte-serializes as a length-prefixed series of bytes represented
    by those hex digits.

    Ex: len(str(Bytes("deadbeef")) == 8   # Eight hex chars
        len(bytes(Bytes("deadbeef")) == 5 # Four data bytes plus varint length

    Implements __json__() method to disambiguate between string and numeric in
    event where hex digits include only numeric digits and no alpha digits.

    """

    def __init__(self, d):
        self.data = d

    def __bytes__(self):
        d = unhexlify(bytes(self.data, "utf-8"))
        return varint(len(d)) + d

    def __json__(self):
        return str(self.data)

    def __str__(self):
        return str(self.data)


class Extension(Array):
    sorted_options = []

    def __init__(self, *args, **kwargs):
        self.json = {}
        a = []
        for key, value in kwargs.items():
            self.json.update({key: value})
        for arg in args:
            if isinstance(arg, dict):
                self.json.update(arg)

        for index, extension in enumerate(self.sorted_options):
            name = extension[0]
            klass = extension[1]
            for key, value in self.json.items():
                if key.lower() == name.lower():
                    a.append(StaticVariant(klass(value), index))
        super().__init__(a)

    def __str__(self):
        """We overload the __str__ function because the json representation is different
        for extensions."""
        return json.dumps(self.json)


# VARINT
def varint(num):
    """
    varint encoding normally saves memory on smaller numbers
    yet retains ability to represent numbers of any magnitude
    specifically, this does the following:
     - take a number, see if it fits in a 7 bit int:
     - - if not, use the 8th as a continuation bit
     - - - continue (see if the number fits in the next 7)
     - - if so, add the last 7 bit chunk
    """
    data = b""
    while num >= 0x80:
        data += bytes([(num & 0x7F) | 0x80])
        num >>= 7
    data += bytes([num])
    return data



r"""
rpc.py

  ____  _ _   ____  _                         
 | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___ 
 |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|
 | |_) | | |_ ___) | | | | (_| | | |  __/\__ \
 |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/
       ____  _             _                  
      / ___|(_) __ _ _ __ (_)_ __   __ _      
      \___ \| |/ _` | '_ \| | '_ \ / _` |     
       ___) | | (_| | | | | | | | | (_| |     
      |____/|_|\__, |_| |_|_|_| |_|\__, |     
               |___/               |___/      


WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

Collection of functions that interact with BitShares nodes using a WebSocket connection.

"""
# DISABLE SELECT PYLINT TESTS
# pylint: disable=broad-except

# STANDARD PYTHON MODULES
import json
import os
import time
from random import shuffle

# GRAPHENE SIGNING MODULES
from bitshares_signing.config import HANDSHAKE_TIMEOUT, NODES, PATH
from bitshares_signing.utilities import read_file, trace, write_file
# THIRD PARTY MODULES
from websocket import create_connection as wss  # handshake to node


def wss_handshake(rpc=None):
    """
    create a wss handshake in less than X seconds, else try again
    """
    shuffle(NODES)
    handshake = HANDSHAKE_TIMEOUT + 1
    while handshake > HANDSHAKE_TIMEOUT:
        try:
            try:
                if rpc is not None:
                    rpc.close()  # attempt to close open stale connection
            except Exception:
                pass
            start = time.time()
            NODES.append(NODES.pop(0))  # rotate list
            node = NODES[0]
            rpc = wss(node, timeout=HANDSHAKE_TIMEOUT)
            handshake = time.time() - start
        except Exception:
            continue
    return rpc


def wss_query(rpc, params, client_order_id=1):
    """
    this definition will place all remote procedure calls (RPC)
    """

    for _ in range(10):
        try:
            # print(it('purple','RPC ' + params[0])('cyan',params[1]))
            # this is the 4 part format of EVERY rpc request
            # params format is ["location", "object", []]
            query = json.dumps(
                {
                    "method": "call",
                    "params": params,
                    "jsonrpc": "2.0",
                    "id": client_order_id,
                }
            )
            # print(query)
            # rpc is the rpc connection created by wss_handshake()
            # we will use this connection to send query and receive json
            rpc.send(query)
            ret = json.loads(rpc.recv())
            try:
                ret = ret["result"]  # if there is result key take it
            except Exception:
                pass
            # print(ret)
            # print('elapsed %.3f sec' % (time.time() - start))
            return ret
        except Exception as error:
            try:  # attempt to terminate the connection
                rpc.close()
            except Exception:
                pass
            trace(error)  # tell me what happened
            # switch NODES
            rpc = wss_handshake(rpc)
            continue


def rpc_block_number(rpc):
    """
    block number and block prefix
    """
    return wss_query(rpc, ["database", "get_dynamic_global_properties", []])


def rpc_account_id(rpc, account_name):
    """
    given an account name return an account id
    """
    ret = wss_query(rpc, ["database", "lookup_accounts", [account_name, 1]])
    return ret[0][1]


def rpc_get_account(rpc, account_name):
    """
    given an account name return an account id
    """
    ret = wss_query(rpc, ["database", "get_account_by_name", [account_name, 1]])
    return ret


def rpc_tx_fees(rpc, account_id):
    # returns fee for limit order create and cancel without 10^precision

    query = [
        "database",
        "get_required_fees",
        [
            [
                ["0", {"from": str(account_id)}],
                ["1", {"from": str(account_id)}],
                ["2", {"from": str(account_id)}],
                ["3", {"from": str(account_id)}],
                ["10", {"from": str(account_id)}],
                ["13", {"from": str(account_id)}],  # add a feed producer
                ["14", {"from": str(account_id)}],
                ["15", {"from": str(account_id)}],
                ["19", {"from": str(account_id)}],  # publish a price feed
                ["47", {"from": str(account_id)}],
                ["59", {"from": str(account_id)}],
                ["61", {"from": str(account_id)}],
                ["63", {"from": str(account_id)}],
            ],
            "1.3.0",
        ],
    ]
    ret = wss_query(rpc, query)
    final_ret = {
        "transfer": ret[0]["amount"],
        "create": ret[1]["amount"],
        "cancel": ret[2]["amount"],
        "call": ret[3]["amount"],
        "asset_create": ret[4]["amount"],
        "add_producer": ret[5]["amount"],
        "issue": ret[6]["amount"],
        "reserve": ret[7]["amount"],
        "publish": ret[8]["amount"],
        "fee_pool": ret[9]["amount"],
        "pool_create": ret[10]["amount"],
        "pool_deposit": ret[11]["amount"],
        "swap": ret[12]["amount"],
    }
    return final_ret


def rpc_ticker(rpc, asset, currency):
    ticker = wss_query(rpc, ["database", "get_ticker", [asset, currency]])
    ret = {}
    ret["base"] = ticker["base"]
    ret["quote"] = ticker["quote"]
    ret["last"] = float(ticker["latest"])
    ret["bid"] = float(ticker["highest_bid"])
    ret["ask"] = float(ticker["lowest_ask"])
    ret["raw"] = ticker
    return ret


def rpc_orderbook(rpc, asset, currency, depth=3):
    """
    Remote procedure call orderbook bids and asks
    ~
    :RPC param str(base): symbol name or ID of the base asset
    :RPC param str(quote): symbol name or ID of the quote asset
    :RPC param int(limit): depth of the order book to retrieve (max limit 50)
    :RPC returns: Order book of the market
    """
    asset_precision = precision(rpc, id_from_name(rpc, asset))
    order_book = wss_query(
        rpc,
        [
            "database",
            "get_order_book",
            [currency, asset, depth],
        ],
    )
    asks = []
    bids = []
    for i, _ in enumerate(order_book["asks"]):
        price = float(order_book["asks"][i]["price"]/10**16)
        if float(price) == 0:
            raise ValueError("zero price in asks")
        volume = float(order_book["asks"][i]["quote"] / 10 ** int(asset_precision))
        asks.append((price, volume))
    for i, _ in enumerate(order_book["bids"]):
        price = float(order_book["bids"][i]["price"]/10**16)
        if float(price) == 0:
            raise ValueError("zero price in bids")
        volume = float(order_book["bids"][i]["quote"] / 10 ** int(asset_precision))
        bids.append((price, volume))
    return {"asks": asks, "bids": bids}


def rpc_pool_book(rpc, pool_id, depth=100):
    # async def gather_orderbook(self, pool_data, rpc, pair, req_params, ws):
    """
    Gather orderbook information either from the pool data or via RPC request
    Parameters:
        pool_data (dict): The data of the pool
        rpc (object): The rpc instance to be used to make the request
        pair (str): The asset pair being queried
        req_params (dict): The request parameters used gather the orderbook

    Returns:
        data (dict): A dictionary containing the bid and ask orderbook information
    """

    def pool(x_start, y_start, delta_x):
        """
        x_start*y_start = k
        x1 = x_start+delta_x
        k / x1 = y1
        y1-y_start = delta_y
        """
        return y_start - (x_start * y_start) / (x_start + delta_x)

    depth += 1

    pool_data = rpc_get_objects(rpc, pool_id)

    balance_a = int(pool_data["balance_a"]) / 10 ** precision(rpc, pool_data["asset_a"])
    balance_b = int(pool_data["balance_b"]) / 10 ** precision(rpc, pool_data["asset_b"])
    konstant = balance_a * balance_b

    # List to store the order book
    bidp, bidv, askp, askv = [], [], [], []
    step = balance_a / depth

    for i in range(1, depth):
        delta_a = i * step
        balance_a2 = balance_a + delta_a
        balance_b2 = konstant / balance_a2
        delta_b = abs(balance_b - balance_b2)
        price = delta_a / delta_b
        # gain = step * price
        askp.append(price)
        askv.append(step)

    for i in range(1, depth):
        delta_a = i * step
        balance_a2 = balance_a - delta_a
        balance_b2 = konstant / balance_a2
        delta_b = abs(balance_b - balance_b2)
        price = delta_a / delta_b
        # gain = step * price
        bidp.append(price)
        bidv.append(step)

    # Sort the order book by price
    asks = list(map(list, sorted(zip(askp, askv), reverse=False)))
    bids = list(map(list, sorted(zip(bidp, bidv), reverse=True)))

    return {"asks": asks, "bids": bids}


def rpc_balances(rpc, account_name):
    """
    account balances
    """
    balances = wss_query(
        rpc,
        [
            "database",
            "get_named_account_balances",
            [account_name, []],
        ],
    )

    # convert from graphene to human-readable
    balances = {
        name_from_id(rpc, obj["asset_id"]): int(obj["amount"]) / 10 ** precision(rpc, obj["asset_id"])
        for obj in balances
    }

    return balances


def rpc_open_orders(rpc, account_name, pair):
    """
    return a list of open orders, for one account, in one market
    """
    ret = wss_query(rpc, ["database", "get_full_accounts", [[account_name], "false"]])
    try:
        limit_orders = ret[0][1]["limit_orders"]
    except Exception:
        limit_orders = []
    market = [pair["currency_id"], pair["asset_id"]]
    orders = []
    for order in limit_orders:
        base_id = order["sell_price"]["base"]["asset_id"]
        quote_id = order["sell_price"]["quote"]["asset_id"]
        if (base_id in market) and (quote_id in market):
            orders.append(order["id"])
    return orders


def rpc_key_reference(rpc, public_key):
    """
    given public key return account id
    """
    return wss_query(rpc, ["database", "get_key_references", [[public_key]]])


def rpc_get_transaction_hex_without_sig(rpc, trx):
    """
    use this to verify the manually serialized trx buffer
    """
    ret = wss_query(rpc, ["database", "get_transaction_hex_without_sig", [trx]])
    return bytes(ret, "utf-8")


def rpc_get_transaction_hex(rpc, trx):
    """
    use this to verify the manually serialized trx buffer
    """
    ret = wss_query(rpc, ["database", "get_transaction_hex", [trx]])
    try:
        return bytes(ret, "utf-8")
    except Exception:
        print(trx)
        print(ret)


def rpc_get_objects(rpc, obj_id):
    """
    use this to verify the manually serialized trx buffer
    """
    return wss_query(rpc, ["database", "get_objects", [[obj_id]]])[0]


def rpc_broadcast_transaction(rpc, trx, client_order_id=1):
    """
    upload the signed transaction to the blockchain
    """
    ret = wss_query(
        rpc, ["network_broadcast", "broadcast_transaction", [trx]], client_order_id
    )

    print(json.dumps(ret, indent=4))

    return ret


def rpc_lookup_asset_symbols(rpc, asset):  # DONE
    """
    Given asset names return asset ids and precisions
    """
    if isinstance(asset, str):
        asset = [asset]

    return wss_query(rpc, ["database", "lookup_asset_symbols", [asset]])


def precision(rpc, object_id):
    """
    Retrieve or fetch and store the precision value for a given object ID.

    Args:
        rpc: RPC connection object
        object_id: Unique identifier of the object

    Returns:
        int: Precision value associated with the object
    """
    # Attempt to load existing precision data from file
    try:
        precs = json.loads(read_file("precisions.txt"))
    except (FileNotFoundError, json.JSONDecodeError):
        precs = {}  # Initialize empty dict if file doesn't exist or is invalid

    # Return cached precision if available
    if object_id in precs:
        return precs[object_id]

    # Fetch precision from RPC and convert to integer
    prec = int(rpc_get_objects(rpc, object_id)["precision"])
    precs[object_id] = prec  # Cache the new precision value

    # Save updated precision data to file
    write_file(os.path.join(PATH, "pipe", "precisions.txt"), json.dumps(precs))
    return prec


def id_from_name(rpc, object_name):
    """
    Get or retrieve and store object ID based on its name.

    Args:
        rpc: RPC connection object
        object_name: Name of the object

    Returns:
        str: Object ID corresponding to the name
    """
    # Load existing name-to-ID mappings
    try:
        precs = json.loads(read_file(os.path.join(PATH, "pipe", "ids_to_names.txt")))
    except (FileNotFoundError, json.JSONDecodeError):
        precs = {}

    # Return cached ID if available
    if object_name in precs:
        return precs[object_name]

    # Fetch ID from RPC using the object name
    prec = rpc_lookup_asset_symbols(rpc, object_name)[0]["id"]
    precs[object_name] = prec  # Cache the result

    # Save updated mappings to file
    write_file(os.path.join(PATH, "pipe", "ids_to_names.txt"), json.dumps(precs))
    return prec


def name_from_id(rpc, obj_id, kind="asset"):
    """
    Retrieve or fetch and store object name based on its ID and type.

    Args:
        rpc: RPC connection object
        obj_id: Object identifier
        kind: Type of object ("asset" or other), defaults to "asset"

    Returns:
        str: Name or symbol of the object
    """
    # Load existing ID-to-name mappings
    try:
        precs = json.loads(read_file(os.path.join(PATH, "pipe", "names_to_ids.txt")))
    except (FileNotFoundError, json.JSONDecodeError):
        precs = {}

    # Create a unique key combining ID and kind
    key = str((obj_id, kind))
    if key in precs:
        return precs[key]

    # Fetch appropriate field based on object kind
    if kind == "asset":
        prec = rpc_get_objects(rpc, obj_id)["symbol"]  # Get asset symbol
    else:
        prec = rpc_get_objects(rpc, obj_id)["name"]  # Get object name

    precs[key] = prec  # Cache the result

    # Save updated mappings to file
    write_file(os.path.join(PATH, "pipe", "names_to_ids.txt"), json.dumps(precs))
    return prec


def is_mpa(rpc, object_id):
    """
    Check if an object is a Market Pegged Asset (MPA).

    Args:
        object_id: Unique identifier of the object

    Returns:
        bool: True if object is an MPA, False otherwise
    """
    # Load existing MPA status data
    try:
        precs = json.loads(read_file(os.path.join(PATH, "pipe", "mpas.txt")))
    except (FileNotFoundError, json.JSONDecodeError):
        precs = {}

    # Return cached status if available
    if object_id in precs:
        return precs[object_id]

    # Check if object has bitasset_data_id field to determine MPA status
    prec = "bitasset_data_id" in rpc_get_objects(rpc, object_id)
    precs[object_id] = prec  # Cache the result

    # Save updated MPA status data to file
    write_file(os.path.join(PATH, "pipe", "mpas.txt"), json.dumps(precs))
    return prec


def unit_test():
    """
    test functionality of select definitions
    """
    rpc = wss_handshake()
    rpc_get_account(rpc, "litepresence1")


if __name__ == "__main__":
    unit_test()
